.. note::

    The metadata value must be UTF-8-encoded and then
    URL-encoded before you include it in the header.
    This is a direct violation of the HTTP/1.1 `basic rules
    <http://www.w3.org/Protocols/rfc2616/rfc2616-sec2.html#sec2.2>`_.
