#!/bin/bash

set -e

cd /home/{{ ansible_user }}/rings

rm -f *.builder *.ring.gz backups/*.builder backups/*.ring.gz

swift-ring-builder object.builder create 10 3 1
swift-ring-builder object.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb1 1
swift-ring-builder object.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb2 1
swift-ring-builder object.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb3 1
swift-ring-builder object.builder rebalance
swift-ring-builder object-1.builder create 10 2 1
swift-ring-builder object-1.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb1 1
swift-ring-builder object-1.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb2 1
swift-ring-builder object-1.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb3 1
swift-ring-builder object-1.builder rebalance
swift-ring-builder object-2.builder create 10 6 1
swift-ring-builder object-2.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb1 1
swift-ring-builder object-2.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb2 1
swift-ring-builder object-2.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb3 1
swift-ring-builder object-2.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb4 1
swift-ring-builder object-2.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb5 1
swift-ring-builder object-2.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb6 1
swift-ring-builder object-2.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb7 1
swift-ring-builder object-2.builder add r1z1-{{ hostvars['object1'].nodepool.private_ipv4 }}:{{ swift_base_port + 10 }}/sdb8 1
swift-ring-builder object-2.builder rebalance
swift-ring-builder container.builder create 10 3 1
swift-ring-builder container.builder add r1z1-{{ hostvars['container1'].nodepool.private_ipv4 }}:{{ swift_base_port + 11 }}/sdb1 1
swift-ring-builder container.builder add r1z1-{{ hostvars['container1'].nodepool.private_ipv4 }}:{{ swift_base_port + 11 }}/sdb2 1
swift-ring-builder container.builder add r1z1-{{ hostvars['container1'].nodepool.private_ipv4 }}:{{ swift_base_port + 11 }}/sdb3 1
swift-ring-builder container.builder rebalance
swift-ring-builder account.builder create 10 3 1
swift-ring-builder account.builder add r1z1-{{ hostvars['account1'].nodepool.private_ipv4 }}:{{ swift_base_port + 12 }}/sdb1 1
swift-ring-builder account.builder add r1z1-{{ hostvars['account1'].nodepool.private_ipv4 }}:{{ swift_base_port + 12 }}/sdb2 1
swift-ring-builder account.builder add r1z1-{{ hostvars['account1'].nodepool.private_ipv4 }}:{{ swift_base_port + 12 }}/sdb3 1
swift-ring-builder account.builder rebalance
