.. note::

    Metadata keys (the name of the metadata) must be treated as case-insensitive
    at all times. These keys can contain ASCII 7-bit characters that are not
    control (0-31) characters, DEL, or a separator character, according to
    `HTTP/1.1 <http://www.w3.org/Protocols/rfc2616/rfc2616.html>`_ .
    The underscore character is silently converted to a hyphen.
